#include<stdio.h>
#include<math.h>
#include<complex.h>

int main(){

  double theta, theta_rad;
  /* 倍精度複素数型の2x2の２次元配列を定義 */
  double _Complex sigmax[2][2], sigmay[2][2], sigmaz[2][2], amatx[2][2], amaty[2][2], amatz[2][2], cmatx[2][2], cmaty[2][2], cmatz[2][2], unit[2][2];
  double _Complex expamatx[2][2], expamaty[2][2], expamatz[2][2];
  int i,j;

  /* 行列指数関数exp(A)を計算する関数のプロトタイプ宣言。第一引数は行列の次元 */
  void expAtaylor(int dim, int, double _Complex [dim][dim], double _Complex [dim][dim]);

  /* sigmax, sigmay, sigmazと単位行列unitに値を代入。Iはcomplex.hで定義された虚数単位 */
  sigmax[0][0] = 0.0;   sigmax[0][1] = 1.0;   sigmax[1][0] = 1.0;   sigmax[1][1] = 0.0;
  sigmay[0][0] = 0.0;   sigmay[0][1] =  -I;   sigmay[1][0] =   I;   sigmay[1][1] = 0.0;
  sigmaz[0][0] = 1.0;   sigmaz[0][1] = 0.0;   sigmaz[1][0] = 0.0;   sigmaz[1][1] =-1.0;
  unit  [0][0] = 1.0;   unit  [0][1] = 0.0;   unit  [1][0] = 0.0;   unit  [1][1] = 1.0; 

  /* 角度は-180度から30度ごとに計算 */
  for(theta =-180.0; theta<181.0; theta+=30){

    /* 角度をラジアンに変換。M_PIはmath.hで定義される円周率 */
    theta_rad = theta * M_PI/180.0;
    printf("theta = %f\n", theta);
    
    /* amat(x,y,z)にi*theta*sigma(x,y,z)を代入 */
    for(i=0;i<2;i++){
      for(j=0;j<2;j++){
	amatx[i][j] = I*theta_rad*sigmax[i][j];
	amaty[i][j] = I*theta_rad*sigmay[i][j];
	amatz[i][j] = I*theta_rad*sigmaz[i][j];	
      }

    }
    
    /* 2x2の行列指数関数をテーラー展開で30次まで計算。入力行列はamat(x,y,z)のアドレスを渡し、出力行列は渡したアドレスexpamat(x,y,z)の配列に上書き */
    expAtaylor(2,30,amatx,expamatx);
    expAtaylor(2,30,amaty,expamaty);
    expAtaylor(2,30,amatz,expamatz);

    /* 行列指数関数の結果の出力 */
    printf("exp(i*theta*sigma)\n");
    for(i=0;i<2;i++){
      for(j=0;j<2;j++){
	printf("%+f%+f i  ", creal(expamatx[i][j]), cimag(expamatx[i][j]));
      }
      printf("        ");
      for(j=0;j<2;j++){
	printf("%+f%+f i  ", creal(expamaty[i][j]), cimag(expamaty[i][j]));
      }
      printf("        ");
      for(j=0;j<2;j++){
	printf("%+f%+f i  ", creal(expamatz[i][j]), cimag(expamatz[i][j]));
      }
      printf("\n");
    }

    /* I cos(theta) + i sigma(x,y,z) sin(theta)をcmat(x,y,z)に代入 */
    for(i=0;i<2;i++){
      for(j=0;j<2;j++){    
	cmatx[i][j] = unit[i][j] * cos(theta_rad) + I * sigmax[i][j]* sin(theta_rad);
	cmaty[i][j] = unit[i][j] * cos(theta_rad) + I * sigmay[i][j]* sin(theta_rad);
	cmatz[i][j] = unit[i][j] * cos(theta_rad) + I * sigmaz[i][j]* sin(theta_rad);
      }
    }

    /* cmat(x,y,z)の出力 */
    printf("I*cos(theta) + i*sigma*sin(theta)\n");
    for(i=0;i<2;i++){
      for(j=0;j<2;j++){
	printf("%+f%+f i  ", creal(cmatx[i][j]), cimag(cmatx[i][j]));
      }
      printf("        ");
      for(j=0;j<2;j++){
	printf("%+f%+f i  ", creal(cmaty[i][j]), cimag(cmaty[i][j]));
      }
      printf("        ");
      for(j=0;j<2;j++){
	printf("%+f%+f i  ", creal(cmatz[i][j]), cimag(cmatz[i][j]));
      }
      printf("\n");
    }
    
  }

  return 0;
  
}


void expAtaylor(int dim, int n, double _Complex amat[dim][dim], double _Complex expamat[dim][dim]){
  /* dimxdim次元の行列のexp(A)をn次までのテーラー展開で計算する関数 
     inputの行列Aはamat(dimxdimの２次元配列)
     outputの行列exp(A)はexpamat(dimxdimの2次元配列)に出力 */

  int i,j,k;
  double _Complex amatk[dim][dim];
  
  /* 行列Aのn乗を計算する関数のプロトタイプ宣言*/
  int powmat(int dim, int, double _Complex [dim][dim], double _Complex[dim][dim]);
  /* 階乗 */
  double factorial(int);

  /* outputの行列expamatの初期化 */
  for(i=0;i<dim;i++){
    for(j=0;j<dim;j++){
      expamat[i][j] = 0.0;
    }
  }
  /* テーラー展開の次数kのループ */
  for(k=0;k<=n;k++){

    /* amat(行列A)のk乗を計算、計算結果(A^k)をamatkに代入 */
    powmat(dim,k,amat,amatk);

    /* テーラー展開の和A^k/k!を計算 */
    for(i=0;i<dim;i++){
      for(j=0;j<dim;j++){
	expamat[i][j] += amatk[i][j]/factorial(k);
      }
    }
  }
  
}

/* dimxdim行列amatのk乗を計算しamatkに代入する関数 */
int powmat(int dim, int k,  double _Complex amat[dim][dim], double _Complex amatk[dim][dim]){

  double _Complex amatkm1[dim][dim];
  int i,j,l;

  /* k<0の場合はエラー */
  if(k < 0){
    printf("k must be zero or positive\n");
    return -1;
  } else if(k==0){
    /* k=0の場合はamatkに単位行列を代入 */
       for(i=0;i<dim;i++){
      for(j=0;j<dim;j++){
	amatk[i][j] = 0.0;
	if(i==j) amatk[i][j] = 1.0;
      }
    }
    return 0;
    
  } else if(k==1){
    /* k=1の場合はamatkはamatとなる */ 
    for(i=0;i<dim;i++){
      for(j=0;j<dim;j++){
	amatk[i][j] = amat[i][j];
      }
    }
    return 0;
  } else {
    /* k>1の場合は再帰的に計算 */

    /* A^{k-1}を計算してamatkm1に代入*/
    powmat(dim,k-1,amat,amatkm1);

    /* amatkの値を初期化 */
    for(i=0;i<dim;i++){
      for(j=0;j<dim;j++){
	amatk[i][j]  = 0.0;
      }
    }

    /* amatkはamat(A)とamatkm1(A^{k-1})の行列積 */
    for(i=0;i<dim;i++){
      for(j=0;j<dim;j++){
	for(l=0;l<dim;l++){
	  amatk[i][j] += amat[i][l] * amatkm1[l][j];
	}
      }
    }
    
  }
    
  return 0;
  
}
  


/* 階乗の計算 */ 
double factorial(int n){
    
  int k;
  double fact;

  fact = 0.0;
  if(n<0) return (fact);

  fact = 1.0;
  if(n==0) return(fact);

  for(k=1;k<=n;k++){

    fact *= (double) k;

  }
    
  return(fact);

}
